-----------------------------------------------------------------------------------------------
-- Client Lua Script for VPtrack
-- Copyright (c) NCsoft. All rights reserved
-- Written By: The Weasel (Entity)
-----------------------------------------------------------------------------------------------
 
require "Window"
 
-----------------------------------------------------------------------------------------------
-- VPtrack Module Definition
-----------------------------------------------------------------------------------------------
local VPtrack = {} 
 
-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
-- e.g. local kiExampleVariableMax = 999
icons = {}

aUnits = {}
aTime = {}
aTimerDiff = {}
iTimeRemaining = 0
Size = 0
fTimer = os.clock()

icons["skull"] = 	{color = {a=1,r=1,g=1,b=1},	sprite = "IconSprites:Icon_Guild_UI_Guild_Skull", 				text = "Debug", 	font = "CRB_Pixel"}
icons["vp"] = 		{color = {a=1,r=1,g=1,b=1},	sprite = "IconSprites:Icon_SkillSpellslinger_void_pact", 		text = nil, 		font = "CRB_Pixel"}

-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function VPtrack:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    -- initialize variables here

    return o
end

function VPtrack:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = {
		-- "UnitOrPackageName",
	}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 

-----------------------------------------------------------------------------------------------
-- VPtrack OnLoad
-----------------------------------------------------------------------------------------------
function VPtrack:OnLoad()
    -- load our form file
	self.xmlDoc = XmlDoc.CreateFromFile("VPtrack.xml")
	Apollo.LoadSprites("WeaselSprite.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
	self.ICO = Apollo.LoadForm(self.xmlDoc, "ICO", nil, self)
end

-----------------------------------------------------------------------------------------------
-- VPtrack OnDocLoaded
-----------------------------------------------------------------------------------------------
function VPtrack:OnDocLoaded()

	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
	    self.wndMain = Apollo.LoadForm(self.xmlDoc, "VPtrackForm", nil, self)
		if self.wndMain == nil then
			Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
			return
		end
		
	    self.wndMain:Show(false, true)
	    self.ICO:Show(true, true)

		-- if the xmlDoc is no longer needed, you should set it to nil
		-- self.xmlDoc = nil
		
		-- Register handlers for events, slash commands and timer, etc.
		-- e.g. Apollo.RegisterEventHandler("KeyDown", "OnKeyDown", self)
		Apollo.RegisterSlashCommand("VPtrack", 	"OnVPtrackOn", 	self)
		Apollo.RegisterSlashCommand("VP", 		"OnVPtrackOn", 	self)
		Apollo.RegisterSlashCommand("vptrack", 	"OnVPtrackOn", 	self)
		Apollo.RegisterSlashCommand("vp", 		"OnVPtrackOn", 	self)

		self.timer = ApolloTimer.Create(0.25, true, "OnTimer", self) --Lookup frequency
		Apollo.RegisterEventHandler("NextFrame", "OnFrame", self) --Draw frequency (NextFrame)

		-- Do additional Addon initialization here
		self.opt = {}
		self:RegisterGUIElements()
		
		-- Load
		if self.OptEnabled 	~= nil then	self.opt.Enabled:SetCheck(self.OptEnabled) 	end
		if self.OptTimer 	~= nil then self.opt.Timer:SetCheck(self.OptTimer) 		end
		if self.OptSize 	~= nil then
			Size = self.OptSize
		else
			Size = 20
		end
		self.opt.Size:SetValue(Size)

		if self.OptDUp 		~= nil then self.opt.DUp:SetCheck(self.OptDUp) 			end
		if self.OptDDown 	~= nil then self.opt.DDown:SetCheck(self.OptDDown) 		end
		if self.OptDLeft 	~= nil then self.opt.DLeft:SetCheck(self.OptDLeft) 		end
		if self.OptDRight 	~= nil then self.opt.DRight:SetCheck(self.OptDRight) 	end

		if not self.opt.DUp:IsChecked() and not self.opt.DDown:IsChecked() and not self.opt.DLeft:IsChecked() and not self.opt.DRight:IsChecked() then
			self.opt.DCenter:SetCheck(true) -- Default option is center
		end
		
		if not self.opt.Enabled:IsChecked() then self:EnableHide() end -- Hide controls after loading settings if main option is not enabled

		ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_Realm, "VPtrack - Tracks void pact in your group. \"/vp\" for options.\n-The Weasel.")
	end
end

-----------------------------------------------------------------------------------------------
-- VPtrack Functions
-----------------------------------------------------------------------------------------------
-- Define general functions here

-- on SlashCommand "/VPtrack"
function VPtrack:OnVPtrackOn()
	self.wndMain:Invoke() -- show the window
end

-- on timer
function VPtrack:OnTimer()
	if self.opt.Enabled:IsChecked() ~= true then return end
	self:CheckVP()
end

function VPtrack:OnFrame()
	self.ICO:DestroyAllPixies()
	if self.opt.Enabled:IsChecked() ~= true then return end
	for key,val in pairs(aUnits) do
		local aScreenPos = GameLib.GetUnitScreenPosition(val)
		if aScreenPos and aScreenPos.bOnScreen and val:IsValid() then
			local ModTime = aTime[key] - (os.clock() - aTimerDiff[key])
			if ModTime < 0 then ModTime = 0 end
			self:CreateIcon(tostring(aScreenPos.nX), tostring(aScreenPos.nY), ModTime, Size)
		end
	end
end

function VPtrack:CheckVP()
	local iMembers = GroupLib.GetMemberCount()
	if iMembers == 0 then
		local CheckUnit = GameLib.GetPlayerUnit() --Not in a party, so only test player unit
		if CheckUnit then
			local ID = CheckUnit:GetId()
			if self:IsBuffOnTarget(CheckUnit, "buff", "name", "Void Pact") then --check for VP buff
				aUnits[ID] 		= CheckUnit 		--Add unit to our table for tracking
				aTime[ID] 		= iTimeRemaining 	--Track time remaining on buff
				aTimerDiff[ID] 	= os.clock() 		--Track time that we logged this to fudge the time remaining without the expensive lookup.
			else
				aUnits[ID] 		= nil 				--Remove unit from our table since they do not have the buff
				aTime[ID] 		= nil 				--Remove unit from our table since they do not have the buff
				aTimerDiff[ID] 	= nil 				--Remove unit from our table since they do not have the buff
			end
		end
		return
	else
		for i=1, GroupLib.GetMemberCount() do --In a party, so iterate through members
			local CheckUnit = GroupLib.GetUnitForGroupMember(i)
				if CheckUnit then
				local ID = CheckUnit:GetId()
				if self:IsBuffOnTarget(CheckUnit, "buff", "name", "Void Pact") then --check for VP buff
					aUnits[ID] 		= CheckUnit 		--Add unit to our table for tracking
					aTime[ID] 		= iTimeRemaining 	--Track time remaining on buff
					aTimerDiff[ID] 	= os.clock() 		--Track time that we logged this to fudge the time remaining without the expensive lookup.
				else
					aUnits[ID] 		= nil 				--Remove unit from our table since they do not have the buff
					aTime[ID] 		= nil 				--Remove unit from our table since they do not have the buff
					aTimerDiff[ID] 	= nil 				--Remove unit from our table since they do not have the buff
				end
			end
		end
	end
end

function VPtrack:CreateIcon(x, y, itime, isize)
	local IcoSize = isize / 2
	local xOffSet = 10
	local yOffSet = 45
	local nIcon = self.ICO:AddPixie({
		strText = icons["vp"].text,
		strFont = icons["vp"].font,
		bLine = false,
		strSprite = icons["vp"].sprite,
		cr = icons["vp"].color,
		loc = {
			fPoints = {0,0,0,0},
			nOffsets = {x-xOffSet-IcoSize, y-yOffSet-IcoSize, x-xOffSet+IcoSize, y-yOffSet+IcoSize}
		},
		flagsText = {
			    DT_CENTER = true,
			    DT_VCENTER = true
		},
	})
	
	--If timer is enabled then continue to draw timer
	if self.opt.Timer:IsChecked() ~= true then return end
	local xextend = 5
	
	--Adjust placement of timer based on selected option (no change on center) 
	if self.opt.DDown:IsChecked() 	then yOffSet = yOffSet - (IcoSize * 2) end
	if self.opt.DUp:IsChecked() 	then yOffSet = yOffSet + (IcoSize * 2) end
	if self.opt.DLeft:IsChecked() 	then xOffSet = xOffSet + (IcoSize * 2) end
	if self.opt.DRight:IsChecked() 	then xOffSet = xOffSet - (IcoSize * 2) end
	
	local nText = self.ICO:AddPixie({
		strText = string.format("%3.1f", tostring(itime)),
		strFont = icons["vp"].font,
		bLine = false,
		strSprite = nil,
		cr = icons["vp"].color,
		loc = {
			fPoints = {0,0,0,0},
			nOffsets = {x-xOffSet-IcoSize-xextend, y-yOffSet-IcoSize, x-xOffSet+IcoSize+xextend, y-yOffSet+IcoSize}
		},
		flagsText = {
			    DT_CENTER = true,
			    DT_VCENTER = true
		},
	})
	
end

function VPtrack:IsBuffOnTarget(ThisUnit, BuffDebuff, NameId, ThisNameId)
	local aCheckBuffs
	local aBuffs = ThisUnit:GetBuffs()
	if BuffDebuff == "buff" then
		aCheckBuffs = aBuffs.arBeneficial
	elseif BuffDebuff == "debuff" then
		aCheckBuffs = aBuffs.arHarmful
	else
		return "error"
	end
	for key,value in pairs(aCheckBuffs) do
		local spell = value.splEffect
		local testme --Testing name or ID?
		if NameId == "id" then
			testme = spell:GetBaseSpellId()
		elseif NameId == "name" then
			testme = spell:GetName()
		else
			return "error"
		end
		if testme == ThisNameId then --Check conditions
			iTimeRemaining = value.fTimeRemaining
			return true 
		end
	end
	return false --Buff/Debuff was NOT found on unit.
end

--Future feature to scale on distance
function VPtrack:CalculateDistance(vector)
	return math.sqrt(math.pow(vector.x, 2) + math.pow(vector.y, 2) + math.pow(vector.z, 2))
end

function VPtrack:DistToLoc(loc, unit)
	local unitPos
	if type(unit) == "table" then
		unitPos = unit
	else
		unitPos = unit:GetPosition()
	end
	local distance = self:CalculateDistance(Vector3.New(loc.x - unitPos.x, loc.y - unitPos.y, loc.z - unitPos.z))
	if distance then 
		return distance
	else
		return nil
	end
end

function VPtrack:OnSave(eType)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then return end
	local tSave = {}
	tSave.Enabled 	= self.opt.Enabled:IsChecked()
	tSave.Timer 	= self.opt.Timer:IsChecked()
	tSave.Size 		= self.opt.Size:GetValue()
	tSave.DUp 		= self.opt.DUp:IsChecked()
	tSave.DDown 	= self.opt.DDown:IsChecked()
	tSave.DLeft 	= self.opt.DLeft:IsChecked()
	tSave.DRight 	= self.opt.DRight:IsChecked()
	tSave.DCenter 	= self.opt.DCenter:IsChecked()
	return tSave
end

function VPtrack:OnRestore(eType, tData)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then return end
	if tData.Enabled 	~= nil then self.OptEnabled 	= tData.Enabled end
	if tData.Timer 		~= nil then self.OptTimer 		= tData.Timer end
	if tData.Size 		~= nil then self.OptSize 		= tData.Size end
	if tData.DUp 		~= nil then self.OptDUp 		= tData.DUp end
	if tData.DDown 		~= nil then self.OptDDown 		= tData.DDown end
	if tData.DLeft 		~= nil then self.OptDLeft 		= tData.DLeft end
	if tData.DRight 	~= nil then self.OptDRight 		= tData.DRight end
	if tData.DCenter 	~= nil then self.OptDCenter 	= tData.DCenter end
end

-----------------------------------------------------------------------------------------------
-- VPtrackForm Functions
-----------------------------------------------------------------------------------------------
function VPtrack:OnOK()
	self:OnSave(GameLib.CodeEnumAddonSaveLevel.Character)
	self.wndMain:Close() -- hide the window
end

function VPtrack:OnCancel()
	self:OnSave(GameLib.CodeEnumAddonSaveLevel.Character)
	self.wndMain:Close() -- hide the window
end

function VPtrack:OnButtonClose( wndHandler, wndControl, eMouseButton )
	self:OnSave(GameLib.CodeEnumAddonSaveLevel.Character)
	self.wndMain:Close() -- hide the window
end

function VPtrack:OnButtonTest1( wndHandler, wndControl, eMouseButton )
	--Debugging
	Print(os.clock())
end


-- UI options
function VPtrack:RegisterGUIElements()
	self.opt.Enabled 	= 	self.wndMain:FindChild("GroupMain"):FindChild("chkEnabled")
	self.opt.Size 		= 	self.wndMain:FindChild("GroupMain"):FindChild("SliderBar")
	self.opt.Timer 		= 	self.wndMain:FindChild("GroupDirection"):FindChild("chkTimer")
	self.opt.DUp 		= 	self.wndMain:FindChild("GroupDirection"):FindChild("ButtonUp")
	self.opt.DDown 		= 	self.wndMain:FindChild("GroupDirection"):FindChild("ButtonDown")
	self.opt.DLeft 		= 	self.wndMain:FindChild("GroupDirection"):FindChild("ButtonLeft")
	self.opt.DRight 	= 	self.wndMain:FindChild("GroupDirection"):FindChild("ButtonRight")
	self.opt.DCenter 	= 	self.wndMain:FindChild("GroupDirection"):FindChild("ButtonCenter")
end

function VPtrack:OnSliderChange( wndHandler, wndControl, fNewValue, fOldValue )
	--Print(fNewValue)
	Size = fNewValue
end

function VPtrack:OnDirectionButton( wndHandler, wndControl, eMouseButton )
	--Print("TEST: " .. wndControl:GetName())
end

function VPtrack:OnButtonEnable( wndHandler, wndControl, eMouseButton )
	if wndControl:IsChecked() then
		self:EnableShow()
	else
		self:EnableHide()
	end
end

function VPtrack:EnableShow()
	self.opt.Size:Show(true)
	self.opt.Timer:Show(true)
	self.opt.DUp:Show(true)
	self.opt.DDown:Show(true)
	self.opt.DLeft:Show(true)
	self.opt.DRight:Show(true)
	self.opt.DCenter:Show(true)
end

function VPtrack:EnableHide()
	self.opt.Size:Show(false)
	self.opt.Timer:Show(false)
	self.opt.DUp:Show(false)
	self.opt.DDown:Show(false)
	self.opt.DLeft:Show(false)
	self.opt.DRight:Show(false)
	self.opt.DCenter:Show(false)
end

-----------------------------------------------------------------------------------------------
-- VPtrack Instance
-----------------------------------------------------------------------------------------------
local VPtrackInst = VPtrack:new()
VPtrackInst:Init()
